# Imports
from itertools import combinations
import time
from turtle import clear
from typing import Deque, Dict, List, Tuple, Set
import numpy as np
import copy
import math
from queue import PriorityQueue
from collections import deque

# Clase Operacion
class Operacion:
    """
    Clase que representa una operacion de un nodo del arbol de busqueda de una instancia
    del JSP cualquiera
    """

    def __init__(self, id: int, maquina: int, peso: int) -> None:
        # Inicializamos los atributos
        self.__id: int = id
        self.__peso: int = peso
        self.__maquina: int = maquina
        self.__predecesores_maquina: Set[Operacion] = set()
        self.__sucesores_maquina: Set[Operacion] = set()
        self.__predecesor_trabajo: Set[Operacion] = set()
        self.__sucesor_trabajo: Set[Operacion] = set()

    # Setters
    def fija_predecesor_trabajo(self, op) -> None:
        self.__predecesor_trabajo.add(op)

    def fija_sucesor_trabajo(self, op) -> None:
        self.__sucesor_trabajo.add(op)

    def anhade_predecesor_maquina(self, op) -> None:
        self.__predecesores_maquina.add(op)

    def anhade_sucesor_maquina(self, op) -> None:
        self.__sucesores_maquina.add(op)

    # Getters
    def id(self) -> int:
        return self.__id

    def peso(self) -> int:
        return self.__peso

    def maquina(self) -> int:
        return self.__maquina

    def predecesor_trabajo(self):
        return self.__predecesor_trabajo

    def sucesor_trabajo(self):
        return self.__sucesor_trabajo

    def predecesores_maquina(self):
        return self.__predecesores_maquina
    
    def sucesores_maquina(self):
        return self.__sucesores_maquina
 
# Clase Nodo
class Nodo:
    """
    Clase que representa un nodo del arbol de busqueda
    """
    def __init__(self, operaciones: List[Operacion], maquinas: List[Operacion]) -> None:
        # Inicializamos los atributos
        self.__operaciones: List[Operacion] = operaciones
        self.__maquinas: List[List[Operacion]] = maquinas
        self.__cabezas: List[int] = np.zeros(len(operaciones))
        self.__colas: List[int] = np.zeros(len(operaciones))

    def operaciones_misma_maquina(self, op: Operacion) -> List[Operacion]:
        # Metodo auxiliar que devuelve las operaciones que comparten maquina con una dada
        if op.id() == 0 or op.id() == len(self.__operaciones) - 1:
            return [op]
        else:
            return self.__maquinas[op.maquina()]

    def cabezas_colas(self, UB: int) -> bool:
        # Realiza el calculo de las cabezas y colas
        # Retorna un booleano que indica si hemos de continuar o no
        continuar: bool = True
        # Calculo de las cabezas
        continuar = self.__propaga_cabezas(0, UB)
        if not continuar:
            return continuar
        # Calculo de las colas
        continuar = self.__propaga_colas(len(self.__operaciones) - 1, UB)
        return continuar

    def __propaga_cabezas(self, op_id: int, UB: int) -> bool:
        # Realiza el calculo de cabezas propagando a partir de una operacion dada
        # Retorna un booleano que indica si se ha de podar o no
        # Inicializamos las variables necesarias
        explorados: Set[int] = set()
        frontera: Deque = deque()
        frontera.append(op_id)
        # Mientras queden operaciones por calcular su cabeza
        while len(frontera) > 0:
            # Obtenemos la operacion actual y se anhade a explorados
            opAct_id: int = frontera.popleft()
            explorados.add(opAct_id)
            opAct: Operacion = self.__operaciones[opAct_id]
            # Calculo de la cabeza
            predecesores_maquina: Set[Operacion] =  opAct.predecesores_maquina()
            maximo: int = 0
            # Maximo subconjuntos
            for i in range(1, len(predecesores_maquina) + 1):
                # Subconjuntos tamanho i
                for sub in combinations(predecesores_maquina, i):
                    min_cabezas: int = math.inf
                    suma: int = 0
                    for pred in sub:
                        suma += pred.peso()
                        min_cabezas = min(min_cabezas, self.__cabezas[pred.id()])
                    maximo = max(maximo, min_cabezas + suma)
            # Maximo con predecesores de trabajo
            for pred in opAct.predecesor_trabajo():
                maximo = max(maximo, self.__cabezas[pred.id()] + pred.peso())
            # Actualizamos el valor
            self.__cabezas[opAct_id] = maximo
            # Comprobacion de una cota inferior
            if self.__cabezas[opAct_id] + opAct.peso() + self.__colas[opAct_id] >= UB :
                return False
            # Anhadimos los sucesores que esten listos a la frontera
            for suc in opAct.sucesores_maquina().union(opAct.sucesor_trabajo()):
                insertar: bool = True
                for pred in suc.predecesores_maquina().union(suc.predecesor_trabajo()):
                    if pred.id() not in explorados:
                        insertar = False
                        break
                if insertar:
                    frontera.append(suc.id())
        # Si no se asigna cabeza al final. Ciclo
        if len(self.__operaciones) - 1 not in explorados:
            return False
        return True

    def __propaga_colas(self, op_id: Operacion, UB: int) -> bool:
        # Realiza el calculo de colas propagando a partir de una operacion dada
        # Retorna un booleano que indica si se ha de podar o no
        # Inicializamos las variables necesarias
        explorados: Set[int] = set()
        frontera: Deque = deque()
        frontera.append(op_id)
        # Mientras queden operaciones por calcular su cola
        while len(frontera) > 0:
            # Obtenemos la operacion actual y se anhade a explorados
            opAct_id: int = frontera.popleft()
            explorados.add(opAct_id)
            opAct: Operacion = self.__operaciones[opAct_id]
            # Calculo cola
            sucesores_maquina: Set[Operacion] =  opAct.sucesores_maquina()
            maximo: int = 0
            # Maximo subconjuntos
            for i in range(1, len(sucesores_maquina) + 1):
                # Subconjuntos tamanho i
                for sub in combinations(sucesores_maquina, i):
                    min_colas: int = math.inf
                    suma: int = 0
                    for suc in sub:
                        suma += suc.peso()
                        min_colas = min(min_colas, self.__colas[suc.id()])
                    maximo = max(maximo, min_colas + suma)
            # Maximo con sucesores de trabajo
            for suc in opAct.sucesor_trabajo():
                maximo = max(maximo, suc.peso() + self.__colas[suc.id()])
            # Actualizamos el valor
            self.__colas[opAct_id] = maximo
            # Comprobacion de una cota inferior
            if self.__cabezas[opAct_id] + opAct.peso() + self.__colas[opAct_id] >= UB :
                return False
            # Anhadimos los predecesores que esten listos a la frontera
            for pred in opAct.predecesores_maquina().union(opAct.predecesor_trabajo()):
                insertar: bool = True
                for suc in pred.sucesores_maquina().union(pred.sucesor_trabajo()):
                    if suc.id() not in explorados:
                        insertar = False
                        break
                if insertar:
                    frontera.append(pred.id())
        return True

    def sol_heur(self):
        # Realiza el calculo de una solucion heuristica alcanzable desde el nodo actual
        # Copiamos el objeto al completo (incluido los objetos referenciados)
        nodo_act: Nodo = copy.deepcopy(self)
        # Inicializamos las variables necesarias
        visitadas: set[Operacion] = set()
        C: set[Operacion] = set()
        C.add(nodo_act.__operaciones[0])
        # MIentras queden operaciones por fijar
        while len(C) != 0:
            # Escogemos u
            u: Operacion
            valor_u: int = math.inf
            for op in C:
                if nodo_act.__cabezas[op.id()] + op.peso() < valor_u:
                    u = op
                    valor_u = nodo_act.__cabezas[op.id()] + op.peso()
            # Calculamos el conjunto conflicto
            C_: set[Operacion] = set()
            for op in C:
                if (u.maquina() == op.maquina()) and (nodo_act.__cabezas[op.id()] < valor_u):
                    C_.add(op)
            # Escogemos c_
            c_: Operacion = None
            # Tamanho 0 cuando es 0 o *. Tamanho 1 si solo esta u
            if len(C_) <= 1:
                c_ = u
            else:
                T: List[Operacion] = [op for op in nodo_act.operaciones_misma_maquina(u) if op not in visitadas]
                cota_c_: int = math.inf
                for op in C_:
                    cota_op: int = nodo_act.__cota_heur(op, T)
                    if cota_op < cota_c_:
                        c_ = op
                        cota_c_ = cota_op
            # PLanificamos c_
            visitadas.add(c_)
            op_maquina_c_: List[Operacion] = nodo_act.operaciones_misma_maquina(c_)
            # Sucesores de c_ en la solucion parcial 
            id_sucesores_c_sol_parcial: List[int] = [op.id() for op in self.__operaciones[c_.id()].sucesores_maquina()]
            # Fijamos las aristas correspondientes
            for op in op_maquina_c_:
                if op not in visitadas and op.id() not in id_sucesores_c_sol_parcial:
                    c_.anhade_sucesor_maquina(op)
                    op.anhade_predecesor_maquina(c_)
                    # no podamos
                    nodo_act.__propaga_cabezas(op.id(), math.inf)
            # Actualizar C
            C.remove(c_)
            sucesores_c_: Set[Operacion] = c_.sucesores_maquina().union(c_.sucesor_trabajo())
            for suc in sucesores_c_:
                anhadir: bool = True
                pred_suc: Set[Operacion] = suc.predecesores_maquina().union(suc.predecesor_trabajo())
                for pred in pred_suc:
                    if pred not in visitadas:
                        anhadir = False
                        break
                if anhadir:
                    C.add(suc)
        return nodo_act
        
    def __cota_heur(self, operacion: Operacion, operaciones_maquina: List[Operacion]) -> int:
        # Retorna el valor de la cota inferior utilizada en el heuristico para la operacion indicada teniendo en cuenta
        # las operaciones pasadas (operaciones procesadas en la misma maquina que todavia no se han fijado)
        # Inicializamos las variables necesarias
        tiempo_procesado: Dict[int, int] = {}
        for op in operaciones_maquina:
            tiempo_procesado[op.id()] = 0
        tiempos_compleccion: Dict[int, int] = {}
        tiempos_compleccion[operacion.id()] = self.__cabezas[operacion.id()] + operacion.peso()
        no_disponibles = PriorityQueue()
        disponibles = PriorityQueue()
        visitadas: Set[Operacion] = set()
        # LLenamos las no disponibles (sin contar la fijada al principio)
        for op in operaciones_maquina:
            if op.id() != operacion.id():
                no_disponibles.put((self.__cabezas[op.id()], op.id()))
        # Tiempo inicio (fin de la operacion)
        tiempo: int = self.__cabezas[operacion.id()] + operacion.peso()
        # Mientras queden por finalizar
        while not (no_disponibles.empty() and disponibles.empty()):
            # Anhadimos las disponibles. Puede que haya mas de una disponible en el mismo tiempo
            while (not no_disponibles.empty()) and (no_disponibles.queue[0][0] <= tiempo):
                oper_id: int = no_disponibles.get()[1]
                disponibles.put((-self.__colas[oper_id], oper_id))
            # Ninguna disponible
            if disponibles.empty():
                tiempo = no_disponibles.queue[0][0]
                continue
            # Obtener aquella con mayor prioridad a las colas. Si hay varias escogemos la que haya sido procesada antes
            candidatas: List[Operacion] = []
            prioridad: int = disponibles.queue[0][0]
            while (not disponibles.empty()) and (disponibles.queue[0][0] == prioridad):
                oper: Operacion = self.__operaciones[disponibles.get()[1]]
                candidatas.append(oper)
            operacion_seleccionada: Operacion = None
            for op in candidatas:
                if op in visitadas:
                    operacion_seleccionada = op
                    break
            # Ninguna ha sido procesada antes
            if operacion_seleccionada is None:
                operacion_seleccionada = candidatas[0]
                visitadas.add(operacion_seleccionada)
            # Devolvemos el resto
            for op in candidatas:
                if op != operacion_seleccionada:
                    disponibles.put((-self.__colas[op.id()], op.id()))
            # Ver si se procesa enetera o no
            siguiente_inicio: int 
            if no_disponibles.empty():
                siguiente_inicio = math.inf
            else:
                siguiente_inicio = no_disponibles.queue[0][0]
            tiempo_max_procesado: int = siguiente_inicio - tiempo
            # Finaliza
            if tiempo_procesado[operacion_seleccionada.id()] + tiempo_max_procesado >= operacion_seleccionada.peso():
                # Actualizamos los tiempos
                tiempos_compleccion[operacion_seleccionada.id()] = tiempo + (operacion_seleccionada.peso() - tiempo_procesado[operacion_seleccionada.id()])
                tiempo += operacion_seleccionada.peso() - tiempo_procesado[operacion_seleccionada.id()]
                tiempo_procesado[operacion_seleccionada.id()] = operacion_seleccionada.peso()
            # No finaliza
            else:
                # Actualizamos los tiempos y se vuelve a anhadir a disponibles
                tiempo_procesado[operacion_seleccionada.id()] += tiempo_max_procesado
                disponibles.put((-self.__colas[operacion_seleccionada.id()], operacion_seleccionada.id()))
                tiempo += tiempo_max_procesado
        # Calculamos la cota inferior
        maximo: int = -math.inf
        for op in operaciones_maquina:
            if  tiempos_compleccion[op.id()] + self.__colas[op.id()] > maximo:
                maximo = tiempos_compleccion[op.id()] + self.__colas[op.id()]
        return maximo

    def camino_critico(self) -> List[Operacion]:
        # Retorna el camino critico del nodo actual
        # Inicializamos la ED donde se almacenaran los pesos de los caminos mas pesados para llegar a cada operacion
        pesos: List[int] = np.zeros(len(self.__operaciones))
        # Actualiza peso tras finalizar operacion
        self.__propaga_pesos(0, pesos)
        # Creamos el camino
        camino: List[int] = []
        # Comenzamos en la operacion *
        op_act: Operacion = self.__operaciones[-1]
        # Mientras no nos encontramos en la operacion 0
        while op_act.id() != 0:
            # Buscamos el predecesor
            pred: Operacion
            peso_pred: int = -math.inf
            predecesores: Set[Operacion] = op_act.predecesores_maquina().union(op_act.predecesor_trabajo())
            for op in predecesores:
                if pesos[op.id()] > peso_pred:
                    pred = op
                    peso_pred = pesos[op.id()]
            # Anhadimos el predecesor al camino
            camino.append(pred)
            # Actualizamos la operacion actual
            op_act = pred
        # Retornamos sin la operacion 0 (el camino se encuentra en orden inverso)
        return camino[:-1]

    def __propaga_pesos(self, op_id: int, pesos: List[int]) -> None:
        # Realiza el calculo de los pesos de los caminos mas pesados para llegar a cada operacion 
        # propagando a partir de una operacion dada
        # Actualiza la variable pesos
        # Inicializamos las variables necesarias
        explorados: Set[int] = set()
        frontera: Deque = deque()
        frontera.append(op_id)
        # Mientras queden operaciones por establecer su peso
        while len(frontera) > 0:
            # Obtenemos la operacion actual
            opAct_id: int = frontera.popleft()
            # Anhadimos la operacion actual a explorados
            explorados.add(opAct_id)
            opAct: Operacion = self.__operaciones[opAct_id]
            # Calculo del peso
            peso_pred: int = 0
            predecesores: Set[Operacion] = opAct.predecesores_maquina().union(opAct.predecesor_trabajo())
            for op in predecesores:
                if pesos[op.id()] > peso_pred:
                    peso_pred = pesos[op.id()]
            # Guardamos el valor
            pesos[opAct.id()] = peso_pred + opAct.peso()
            # Revisamos los sucesores para posibles inserciones
            for suc in opAct.sucesores_maquina().union(opAct.sucesor_trabajo()):
                insertar: bool = True
                for pred in suc.predecesores_maquina().union(suc.predecesor_trabajo()):
                    if pred.id() not in explorados:
                        insertar = False
                        break
                if insertar:
                    frontera.append(suc.id())
        return
        
    def peso_camino(self, camino: List[Operacion]) -> int:
        # Retorna el peso del camino pasado como parametro
        # Inicializamos la variable que almacenara el peso acumulado
        peso: int = 0
        # Anhadimos el peso de cada operacion
        for op in camino:
            peso += op.peso()
        # Retornamos el peso total
        return peso

    def bloques(self, camino: List[Operacion]) -> List[List[Operacion]]:
        # Retorna las operaciones del camino pasado como parametro agrupadas en bloques
        # El camino es dado en orden inverso
        # Trabajamos con el id para pasar del camino que contiene las operaciones de la solucion heuristica
        # a las operaciones de este nodo
        # Inicializamos las variables necesarias
        bloques: List[List[Operacion]] = []
        bloques.append([self.__operaciones[camino[-1].id()]])
        # Recorremos el camino. Se toman los indices de esta manera para que sea recorrido de manera inversa
        for i in range(2, len(camino) + 1):
            op_act: Operacion = camino[-i]
            op_ant: Operacion = camino[-i + 1]
            # Si la operacion comparte maquina con la anterior
            if op_act.maquina() == op_ant.maquina():
                # Se anhade al mismo bloque
                bloques[-1].append(self.__operaciones[op_act.id()])
            else:
                # Se anhade a un nuevo bloque
                bloques.append([self.__operaciones[op_act.id()]])
        # Nos quedamos solo con los bloques. Eliminamos los agrupamientos de una sola operacion
        bloques = [bloque for bloque in bloques if len(bloque) > 1]
        # Retornamos los bloques
        return bloques
    
    def conjuntos_ramificaion(self, bloques: List[List[Operacion]], camino: List[Operacion]) -> List[Tuple[int, int, List[Operacion]]]:
        # Retorna los conjuntos de ramificacion dado el camino critico y los bloques del mismo
        # Estos se identifican por tres variables. La primera indica el bloque.
        # La segunda indica si se trata de un conjunto before (0) o after (1). La ultima se trata de una lista que contiene
        # las operaciones del conjunto
        # Observamos si es necesario anhadir el conjunto before del primer bloque o el conjunto after del ultimo
        # Trabajamos con los id. El camino posee las operaciones de la solucion heuristica y el bloque las del nodo actual
        primer_bloque: List[Operacion] = bloques[0]
        anhadir_before_primero: bool = primer_bloque[0].id() != camino[-1].id()
        ultimo_bloque: List[Operacion] = bloques[-1]
        anhadir_after_ultimo: bool = ultimo_bloque[-1].id() != camino[0].id()
        # Ordenamos los bloques segun la longitud
        bloques.sort(key = len, reverse = True)
        # Calculamos los conjuntos
        conjuntos: List[Tuple[int, int, List[Operacion]]] = []
        # Para cada bloque
        for i in range(len(bloques)):
            # Creamos los conjuntos before y after para el bloque
            conjunto_before: List[int, int, List[Operacion]] = [i, 0, []]
            conjunto_after: List[int, int, List[Operacion]] = [i, 1, []]
            # Anhadimos las operaciones correspondientes a cada bloque. Se revisan los ciclos de longitud 2
            for j in range(len(bloques[i])):
                if (j != 0) and (len(bloques[i][j].predecesores_maquina().intersection(bloques[i])) == 0):
                    conjunto_before[2].append(bloques[i][j])
                if (j != len(bloques[i]) - 1) and (len(bloques[i][j].sucesores_maquina().intersection(bloques[i])) == 0):
                    conjunto_after[2].append(bloques[i][j])
            # Anhadimos los conjuntos teniendo en cuenta si es posible prescindir de alguno
            if bloques[i] != primer_bloque or anhadir_before_primero:
                conjuntos.append(conjunto_before)
            if bloques[i] != ultimo_bloque or anhadir_after_ultimo:
                conjuntos.append(conjunto_after)
        return conjuntos

    def cota_inf(self, op: Operacion, bloque: List[Operacion], pos: int) -> int:
        # Retorna el valor de la cota inferior del sucesor resultante de establecer la operacion pasada como parametro
        # al inicio o al final del bloque pasado como parametro tambien
        # La variable pos indica si la operacion ha de establecerse al inicio o al final del bloque
        # Pos 0 before (inicio), Pos 1 after (final)
        # Si se establace al inicio
        if pos == 0:
            # Inicializamos los valores necesarios para la cota
            maximo: int = -math.inf
            suma: int = 0
            minimo_colas: int = math.inf
            # Recorremos las operaciones del bloque distintas de la que hemos de fijar para calcular dichos valores
            for op_bloque in bloque:
                if op != op_bloque:
                    suma += op_bloque.peso()
                    minimo_colas = min(minimo_colas, self.__colas[op_bloque.id()])
                    maximo = max(maximo, op_bloque.peso() + self.__colas[op_bloque.id()])
            # Retornamos el valor de la cota inferior
            return self.__cabezas[op.id()] + op.peso() + max(maximo, suma + minimo_colas)
        # Si se establace al final
        if pos == 1:
            # Inicializamos los valores necesarios para la cota
            maximo: int = -math.inf
            suma: int = 0
            minimo_cabezas: int = math.inf
            # Recorremos las operaciones del bloque distintas de la que hemos de fijar para calcular dichos valores
            for op_bloque in bloque:
                if op != op_bloque:
                    suma += op_bloque.peso()
                    minimo_cabezas = min(minimo_cabezas, self.__cabezas[op_bloque.id()])
                    maximo = max(maximo, self.__cabezas[op_bloque.id()] + op_bloque.peso())
            # Retornamos el valor de la cota inferior
            return  max(maximo, minimo_cabezas + suma) + op.peso() + self.__colas[op.id()]

    def select(self, numMaquina: int, UB: int) -> bool:
        # Metodo que permite fijar todas las aristas directas relativas a una maquina
        # Retorna un booleano que indica si hemos fijado alguna arista nueva
        fijado: bool = False
        operaciones_maquina: List[Operacion] = self.__maquinas[numMaquina]
        # Se revisan todos los pares ordenados de operaciones distintas
        for op1 in operaciones_maquina:
            for op2 in operaciones_maquina:
                # Comprobamos si la arista puede ser fijada y no lo ha sido previamente
                if (op1 != op2) and (op1 not in op2.sucesores_maquina()) and (op2 not in op1.predecesores_maquina()) and (self.__cabezas[op1.id()] + op1.peso() + op2.peso() + self.__colas[op2.id()] >= UB):
                    # Fijamos la arista y actualizamos le booleano
                    fijado = True
                    op2.anhade_sucesor_maquina(op1)
                    op1.anhade_predecesor_maquina(op2)
        return fijado

    def calculo_JPS(self, operaciones_maquina: List[Operacion], UB):
        # Retorna el JPS para las operaciones pasadas. Se identifica por tres variables.
        # La primera se trata de un booleano que indica si se ha podado o no. La segunda se trata de un registro del JPS
        # (en que instante que operacion ha sido procesada y durante cuento tiempo). La ultima almacena los tiempos de compleccion
        # Inicializamos las ED necesarias
        registro: List[List[int]] = []
        tiempo_procesado: Dict[int, int] = {}
        for op in operaciones_maquina:
            tiempo_procesado[op.id()] = 0
        tiempos_compleccion: Dict[int, int] = {}
        no_disponibles = PriorityQueue()
        disponibles = PriorityQueue()
        visitadas: Set[Operacion] = set()
        # LLenamos las no disponibles
        for op in operaciones_maquina:
            no_disponibles.put((self.__cabezas[op.id()], op.id()))
        # Tiempo inicio
        tiempo: int = no_disponibles.queue[0][0]
        # Mientras queden operaciones por finalizar
        while not (no_disponibles.empty() and disponibles.empty()):
            # Anhadimos las disponibles. Puede que haya mas de una disponible en el mismo tiempo
            while (not no_disponibles.empty()) and (no_disponibles.queue[0][0] == tiempo):
                oper_id: int = no_disponibles.get()[1]
                # Valor negativo en las colas para que la prioridad sea la correcta
                disponibles.put((-self.__colas[oper_id], oper_id))
            # Ninguna disponible
            if disponibles.empty():
                tiempo = no_disponibles.queue[0][0]
                continue
            # Obtener aquella con mayor prioridad en las colas. Si hay varias escogemos la que haya sido procesada antes
            candidatas: List[Operacion] = []
            prioridad: int = disponibles.queue[0][0]
            while (not disponibles.empty()) and (disponibles.queue[0][0] == prioridad):
                oper: Operacion = self.__operaciones[disponibles.get()[1]]
                candidatas.append(oper)
            operacion_seleccionada: Operacion = None
            for op in candidatas:
                if op in visitadas:
                    operacion_seleccionada = op
                    break
            # Ninguna ha sido procesada antes
            if operacion_seleccionada is None:
                operacion_seleccionada = candidatas[0]
                visitadas.add(operacion_seleccionada)
            # Devolvemos el resto
            for op in candidatas:
                if op != operacion_seleccionada:
                    disponibles.put((-self.__colas[op.id()], op.id()))
            # Ver si se procesa enetera o no
            siguiente_inicio: int 
            if no_disponibles.empty():
                siguiente_inicio = math.inf
            else:
                siguiente_inicio = no_disponibles.queue[0][0]
            tiempo_max_procesado: int = siguiente_inicio - tiempo
            # Finaliza
            if tiempo_procesado[operacion_seleccionada.id()] + tiempo_max_procesado >= operacion_seleccionada.peso():
                # Actualizamos los tiempos y las ED
                registro.append([tiempo, operacion_seleccionada.id(), operacion_seleccionada.peso() - tiempo_procesado[operacion_seleccionada.id()]])
                tiempos_compleccion[operacion_seleccionada.id()] = tiempo + (operacion_seleccionada.peso() - tiempo_procesado[operacion_seleccionada.id()])
                tiempo += operacion_seleccionada.peso() - tiempo_procesado[operacion_seleccionada.id()]
                tiempo_procesado[operacion_seleccionada.id()] = operacion_seleccionada.peso()
            # No finaliza
            else:
                # Actualizamos los tiempos y las ED. Insertamos la operacion en sisponibles
                registro.append([tiempo, operacion_seleccionada.id(), tiempo_max_procesado])
                tiempo_procesado[operacion_seleccionada.id()] += tiempo_max_procesado
                disponibles.put((-self.__colas[operacion_seleccionada.id()], operacion_seleccionada.id()))
                tiempo += tiempo_max_procesado
        # Comprobamos la cota inferior
        maximo: int = -math.inf
        for op in operaciones_maquina:
            if tiempos_compleccion[op.id()] + self.__colas[op.id()] > maximo:
                maximo = tiempos_compleccion[op.id()] + self.__colas[op.id()]
        if maximo >= UB:
            return [True, registro, tiempos_compleccion] 
        return [False, registro, tiempos_compleccion]

    def calculo_DJPS(self, operaciones_maquina: List[Operacion], UB):
        # Retorna el DJPS para las operaciones pasadas. Se identifica por tres variables.
        # La primera se trata de un booleano que indica si se ha podado o no. La segunda se trata de un registro del JPS
        # (en que instante que operacion ha sido procesada y durante cuento tiempo). La ultima almacena los tiempos de compleccion
        # Inicializamos las ED necesarias
        registro: List[List[int]] = []
        tiempo_procesado: Dict[int, int] = {}
        for op in operaciones_maquina:
            tiempo_procesado[op.id()] = 0
        tiempos_compleccion: Dict[int, int] = {}
        no_disponibles = PriorityQueue()
        disponibles = PriorityQueue()
        visitadas: Set[Operacion] = set()
        # LLenamos las no disponibles
        for op in operaciones_maquina:
            no_disponibles.put((self.__colas[op.id()], op.id()))
        # Tiempo inicio
        tiempo: int = no_disponibles.queue[0][0]
        # Mientras queden operaciones por finalizar
        while not (no_disponibles.empty() and disponibles.empty()):
            # Anhadimos las disponibles. Puede que haya mas de una disponible en el mismo tiempo
            while (not no_disponibles.empty()) and (no_disponibles.queue[0][0] == tiempo):
                oper_id: int = no_disponibles.get()[1]
                # Valor negativo en las cabezas para que la prioridad sea la correcta
                disponibles.put((-self.__cabezas[oper_id], oper_id))
            # Ninguna disponible
            if disponibles.empty():
                tiempo = no_disponibles.queue[0][0]
                continue
            # Obtener aquella con mayor prioridad en las cabezas. Si hay varias escogemos la que haya sido procesada antes
            candidatas: List[Operacion] = []
            prioridad: int = disponibles.queue[0][0]
            while (not disponibles.empty()) and (disponibles.queue[0][0] == prioridad):
                oper: Operacion = self.__operaciones[disponibles.get()[1]]
                candidatas.append(oper)
            operacion_seleccionada: Operacion = None
            for op in candidatas:
                if op in visitadas:
                    operacion_seleccionada = op
                    break
            # Ninguna ha sido procesada antes
            if operacion_seleccionada is None:
                operacion_seleccionada = candidatas[0]
                visitadas.add(operacion_seleccionada)
            # Devolvemos el resto
            for op in candidatas:
                if op != operacion_seleccionada:
                    disponibles.put((-self.__cabezas[op.id()], op.id()))
            # Ver si se procesa enetera o no
            siguiente_inicio: int 
            if no_disponibles.empty():
                siguiente_inicio = math.inf
            else:
                siguiente_inicio = no_disponibles.queue[0][0]
            tiempo_max_procesado: int = siguiente_inicio - tiempo
            # Finaliza
            if tiempo_procesado[operacion_seleccionada.id()] + tiempo_max_procesado >= operacion_seleccionada.peso():
                # Actualizamos los tiempos y las ED
                registro.append([tiempo, operacion_seleccionada.id(), operacion_seleccionada.peso() - tiempo_procesado[operacion_seleccionada.id()]])
                tiempos_compleccion[operacion_seleccionada.id()] = tiempo + (operacion_seleccionada.peso() - tiempo_procesado[operacion_seleccionada.id()])
                tiempo += operacion_seleccionada.peso() - tiempo_procesado[operacion_seleccionada.id()]
                tiempo_procesado[operacion_seleccionada.id()] = operacion_seleccionada.peso()
            # No finaliza
            else:
                # Actualizamos los tiempos y las ED. Insertamos la operacion en sisponibles
                registro.append([tiempo, operacion_seleccionada.id(), tiempo_max_procesado])
                tiempo_procesado[operacion_seleccionada.id()] += tiempo_max_procesado
                disponibles.put((-self.__cabezas[operacion_seleccionada.id()], operacion_seleccionada.id()))
                tiempo += tiempo_max_procesado
        # Comprobamos la cota inferior
        maximo: int = -math.inf
        for op in operaciones_maquina:
            if tiempos_compleccion[op.id()] + self.__cabezas[op.id()] > maximo:
                maximo = tiempos_compleccion[op.id()] + self.__cabezas[op.id()]
        if maximo >= UB:
            return [True, registro, tiempos_compleccion] 
        return [False, registro, tiempos_compleccion]


    def calcula_cabeza_adicional(self, c: Operacion, registro: List[List[int]], tiempos_complecion: Dict[int, int], UB) -> None:
        # Metodo que calcula (si es posible) una nueva cabeza (relativa al metodo para fijar disyunciones adicionales) de la 
        # operacion pasada como parametro teniendo en cuenta la informacion del JPS pasada como parametro tambien
        # Inicializamos las ED necesarias
        operaciones_maquina: List[Operacion] = self.__maquinas[c.maquina()]
        # Calculamos el tiempo procesado antes de rc gracias al registro
        tiempo_procesado_antes_rc: Dict[int, int] = {}
        for op in operaciones_maquina:
            tiempo_procesado_antes_rc[op.id()] = 0
        for evento in registro:
            if evento[0] == self.__cabezas[c.id()]:
                break
            tiempo_procesado_antes_rc[evento[1]] += evento[2]
        # Calculo del conjunto K_c +
        kc_plus: List[Operacion] = []
        for op in operaciones_maquina:
            if (tiempos_complecion[op.id()] > self.__cabezas[c.id()]) and (self.__colas[op.id()] > self.__colas[c.id()]):
                kc_plus.append(op)
        # Si es vacio no es posible calcular la nueva cabeza
        if len(kc_plus) == 0:
            return
        # Se ordena  K_c + para poder calcular  K_c *
        kc_plus.sort(key = self.orden_cola)
        # Calculo del conjunto K_c *
        sumatorio: int = sum([op.peso() - tiempo_procesado_antes_rc[op.id()] for op in kc_plus])
        kc_star: List[Operacion] = None
        for i in range(len(kc_plus)):
            if self.__cabezas[c.id()] + c.peso() + sumatorio + self.__colas[kc_plus[i].id()] >= UB:
                kc_star = kc_plus[i:]
                break
            sumatorio -= kc_plus[i].peso() - tiempo_procesado_antes_rc[kc_plus[i].id()]
        # Si es vacio no es posible calcular la nueva cabeza
        if kc_star is None:
            return
        # Calculo rj_c
        rj_c: int = -math.inf
        for op in kc_star:
            if tiempos_complecion[op.id()] > rj_c:
                rj_c = tiempos_complecion[op.id()]
        # Actualizamos cabeza
        self.__cabezas[c.id()] = rj_c
        return

    def calcula_cola_adicional(self, c: Operacion, registro: List[List[int]], tiempos_complecion: Dict[int, int], UB) -> None:
        # Metodo que calcula (si es posible) una nueva cola (relativa al metodo para fijar disyunciones adicionales) de la 
        # operacion pasada como parametro teniendo en cuenta la informacion del DJPS pasada como parametro tambien
        # Inicializamos las ED necesarias
        operaciones_maquina: List[Operacion] = self.__maquinas[c.maquina()]
        # Calculamos el tiempo procesado antes de qc gracias al registro
        tiempo_procesado_antes_qc: Dict[int, int] = {}
        for op in operaciones_maquina:
            tiempo_procesado_antes_qc[op.id()] = 0
        for evento in registro:
            if evento[0] == self.__colas[c.id()]:
                break
            tiempo_procesado_antes_qc[evento[1]] += evento[2]
        # Calculo del conjunto D_c +
        dc_plus: List[Operacion] = []
        for op in operaciones_maquina:
            if (tiempos_complecion[op.id()] > self.__colas[c.id()]) and (self.__cabezas[op.id()] > self.__cabezas[c.id()]):
                dc_plus.append(op)
        # Si es vacio no es posible calcular la nueva cola cabeza
        if len(dc_plus) == 0:
            return
        # Se ordena  D_c + para poder calcular  D_c *
        dc_plus.sort(key = self.orden_cabeza)
        # Calculo del conjunto D_c *
        sumatorio: int = sum([op.peso() - tiempo_procesado_antes_qc[op.id()] for op in dc_plus])
        dc_star: List[Operacion] = None
        for i in range(len(dc_plus)):
            if self.__cabezas[dc_plus[i].id()]+  sumatorio + c.peso() + self.__colas[c.id()] >= UB:
                dc_star = dc_plus[i:]
                break
            sumatorio -= dc_plus[i].peso() - tiempo_procesado_antes_qc[dc_plus[i].id()]
        # Si es vacio no es posible calcular la nueva cola
        if dc_star is None:
            return
        # Calculo qj_c
        qj_c: int = -math.inf
        for op in dc_star:
            if tiempos_complecion[op.id()] > qj_c:
                qj_c = tiempos_complecion[op.id()]
        # Actualizamos cola
        self.__colas[c.id()] = qj_c
        return
    
    def metodo_adicional(self, UB: int) -> bool:
        # Metodo para fijar disyunciones adicionales
        # Retorna un booleano que indica si continuamos o no (poda)
        # Inicializamos las ED necesarias
        continuar: bool = True
        fijado_general: bool
        poda: bool 
        registro: List[List[int]]
        tiempos_complecion: Dict[int, int]
        # Mientras podamos seguir fijando disyunciones
        while True:
            # Durante esta iteracion todavia no se ha fijado ninguna arista
            fijado_general = False
            # Calculo arcos primales
            # Para cada maquina
            for i in range(len(self.__maquinas)):
                # Calculo del JPS para las operaciones de la maquina
                op_maquina: List[Operacion] = self.__maquinas[i]
                poda, registro, tiempos_complecion = self.calculo_JPS(op_maquina, UB)
                # Si se poda
                if poda:
                    continuar = False
                    return continuar
                # Calculo de las nuevas cabezas para las operaciones de la maquina
                for op in op_maquina:
                    self.calcula_cabeza_adicional(op, registro, tiempos_complecion, UB)
                # Si al pasar el metodo select para la maquina actual fijamos alguna disyuncion
                if self.select(i, UB):
                    fijado_general = True
            # Calculo de nuevas cabezas y colas si es necesario
            if fijado_general:
                continuar = self.cabezas_colas(UB)
                if not continuar:
                    return continuar 
            # Calculo arcos duales
            # Para cada maquina
            for i in range(len(self.__maquinas)):
                # Calculo del DJPS para las operaciones de la maquina
                op_maquina: List[Operacion] = self.__maquinas[i]
                poda, registro, tiempos_complecion = self.calculo_DJPS(op_maquina, UB)
                # Si se poda
                if poda:
                    continuar = False
                    return continuar
                # Calculo de las nuevas colas para las operaciones de la maquina
                for op in op_maquina:
                    self.calcula_cola_adicional(op, registro, tiempos_complecion, UB)
                # Si al pasar el metodo select para la maquina actual fijamos alguna disyuncion
                if self.select(i, UB):
                    fijado_general = True
            # Si durante la iteracion no se ha fijado ninguna disyuncion
            if not fijado_general:
                break
            # Calculo de nuevas cabezas y colas
            continuar = self.cabezas_colas(UB)
            if not continuar:
                return continuar 
        return continuar

    def operaciones(self):
        # Retorna las operaciones del nodo en el que nos encontramos
        return self.__operaciones

    def orden_cabeza(self, op) -> int:
        # Retorna la cabeza de una operacion dada. Permite ordenar segun los valores de las cabezas
        return self.__cabezas[op.id()]

    def orden_cola(self, op) -> int:
        # Retorna la cola de una operacion dada. Permite ordenar segun los valores de las colas
        return self.__colas[op.id()]

# Clase JSP
class ProblemaJSP:
    """
    Clase que representa una instancia del JPS
    """
    def __init__(self, operaciones_input: List[List[int]]) -> None:
        # Inicializamos los atributos utilizando la informacion pasada como parametro
        self.__operaciones: List[Operacion] = []
        numOperaciones: int = 0
        # Calculamos en numero de operaciones
        for trabajo in operaciones_input:
            numOperaciones += len(trabajo) // 2
        # Creamos las operaciones 0 y *
        op0: Operacion = Operacion(0, None, 0)
        self.__operaciones.append(op0)
        opN: Operacion = Operacion(numOperaciones + 1, None, 0)
        # Creamos el resto de operaciones fijando las aristas relativas a restricciones de trabajo
        id_act: int = 0
        for trabajo in operaciones_input:
            op_trabajo: List[Operacion] = []
            for i in range(len(trabajo) // 2):
                id_act += 1
                op_trabajo.append(Operacion(id_act, trabajo[2 * i], trabajo[2 * i + 1]))
            op0.fija_sucesor_trabajo(op_trabajo[0])
            op_trabajo[0].fija_predecesor_trabajo(op0)
            opN.fija_predecesor_trabajo(op_trabajo[-1])
            op_trabajo[-1].fija_sucesor_trabajo(opN)
            for i in range(len(op_trabajo) - 1):
                op_trabajo[i].fija_sucesor_trabajo(op_trabajo[i+1])
                op_trabajo[i+1].fija_predecesor_trabajo(op_trabajo[i])
            self.__operaciones += op_trabajo
        self.__operaciones.append(opN)
        # Guardamos las operaciones agrupadas por maquinas
        num_maquinas: int = 0
        for op in self.__operaciones:
            if (op.maquina() is not None) and (op.maquina() + 1 > num_maquinas):
                num_maquinas = op.maquina() + 1
        self.__maquinas: List[List[Operacion]] = [[] for _ in range(num_maquinas)]
        for op in self.__operaciones:
            if op.maquina() is not None:
                self.__maquinas[op.maquina()].append(op)
        # Inicializamos el numero de nodos expandidos
        self.__numNodosExpandidos: int = 0
        
    def numNodosExpandidos(self) -> int:
        # Retorna el numero de nodos expandidos en una ejecucion del algoritmo
        return self.__numNodosExpandidos

    def solucion_optima_rec(self):
        # Retorna la solucion optima del problema
        # Inicializamos las variables necesarias para el algoritmo
        mejor_solucion: List[Nodo, int]
        nodo_inicial: Nodo = Nodo(self.__operaciones, self.__maquinas)
        mejor_solucion = [nodo_inicial, math.inf]
        # Reiniciamos los nodos expandidos
        self.__numNodosExpandidos = 0
        # Comenzamos la ejecucion
        self.__branch_and_bound(nodo_inicial, mejor_solucion)
        # Retornamos la mejor solucion
        return mejor_solucion

    def __branch_and_bound(self, nodo: Nodo, mejor_solucion) -> None:
        # Aumentamos el numero de nodos expandidos
        self.__numNodosExpandidos += 1
        # Booleano auxiliar para la poda
        continuar: bool = True
        # Calculo de cabezas y colas
        continuar = nodo.cabezas_colas(mejor_solucion[1])
        if not continuar:
            return
        # Metodo para fijar disyunciones adicionales
        continuar = nodo.metodo_adicional(mejor_solucion[1])
        if not continuar:
            return
        # Calculo de la solucion heurustica, su camino critico y el peso de este
        sol_heur: Nodo = nodo.sol_heur()
        camino_critico: List[Operacion] = sol_heur.camino_critico()
        sol_heur_cmax: int = sol_heur.peso_camino(camino_critico)
        # Si la solucion obtenida es mejor actualizamos
        if sol_heur_cmax < mejor_solucion[1]:
            mejor_solucion[0] = sol_heur
            mejor_solucion[1] = sol_heur_cmax
            print(sol_heur_cmax)
        # Calculo de los bloques
        bloques: List[List[Operacion]] = nodo.bloques(camino_critico)
        # Si no hay mejores soluciones
        if len(bloques) == 0:
            return
        # Calculo de los conjuntos de ramificacion
        conjuntos_ramificacion: List[Tuple[int, int, List[Operacion]]] = nodo.conjuntos_ramificaion(bloques, camino_critico)
        # Comenzamos con la ramificacion
        for conjunto in conjuntos_ramificacion:
            # Ordenamos las operaciones
            op_ordenadas: List[Operacion]
            if conjunto[1] == 0:
                op_ordenadas = sorted(conjunto[2], key = nodo.orden_cabeza)
            if conjunto[1] == 1:
                op_ordenadas = sorted(conjunto[2], key = nodo.orden_cola)
            # Creamos los sucesores
            sucesor: Nodo
            for op in op_ordenadas:
                # Si es Before
                if conjunto[1] == 0:
                    sucesor = copy.deepcopy(nodo)
                    for op_bloque in bloques[conjunto[0]]:
                        if op_bloque != op:
                            # Las fijamos en el sucesor
                            sucesor.operaciones()[op.id()].anhade_sucesor_maquina(sucesor.operaciones()[op_bloque.id()])
                            sucesor.operaciones()[op_bloque.id()].anhade_predecesor_maquina(sucesor.operaciones()[op.id()])
                # Si es After
                if conjunto[1] == 1:
                    sucesor = copy.deepcopy(nodo)
                    for op_bloque in bloques[conjunto[0]]:
                        if op_bloque != op:
                            # Las fijamos en el sucesor
                            sucesor.operaciones()[op.id()].anhade_predecesor_maquina(sucesor.operaciones()[op_bloque.id()])
                            sucesor.operaciones()[op_bloque.id()].anhade_sucesor_maquina(sucesor.operaciones()[op.id()])
                # Comprobamos que cumplen la cota
                if nodo.cota_inf(op, bloques[conjunto[0]], conjunto[1]) < mejor_solucion[1]:
                    self.__branch_and_bound(sucesor, mejor_solucion)
            # Una vez inspeccionado el conjuto podemos fijar otras aristas
            # Si es Before
            if conjunto[1] == 0:
                primera_op: Operacion = bloques[conjunto[0]][0]
                for op_bloque in bloques[conjunto[0]]:
                    if primera_op != op_bloque:
                        # Las fijamos en el nodo. De esta forma las tienen los sucesores
                        nodo.operaciones()[primera_op.id()].anhade_sucesor_maquina(nodo.operaciones()[op_bloque.id()])
                        nodo.operaciones()[op_bloque.id()].anhade_predecesor_maquina(nodo.operaciones()[primera_op.id()])
            # Si es after
            if conjunto[1] == 1:
                ultima_op: Operacion = bloques[conjunto[0]][-1]
                for op_bloque in bloques[conjunto[0]]:
                    if ultima_op != op_bloque:
                        # Las fijamos en el nodo. De esta forma las tienen los sucesores
                        nodo.operaciones()[op_bloque.id()].anhade_sucesor_maquina(nodo.operaciones()[ultima_op.id()])
                        nodo.operaciones()[ultima_op.id()].anhade_predecesor_maquina(nodo.operaciones()[op_bloque.id()])
        return

    def solucion_optima_iter(self):
        # Retorna la solucion optima del problema
        # Reiniciamos los nodos expandidos
        self.__numNodosExpandidos = 0
        # Inicializamos las variables necesarias para el algoritmo
        nodo_inicial: Nodo = Nodo(self.__operaciones, self.__maquinas)
        mejor_solucion: List[Nodo, int] = [nodo_inicial, math.inf]
        frontera: List[Nodo] = []
        frontera .append(nodo_inicial)
        # Comenzamos la busqueda
        while len(frontera) > 0:
            # Aumentamos el numero de nodos expandidos
            self.__numNodosExpandidos += 1
            # Obtenemos el nodo actual
            nodo_act: Nodo = frontera.pop()
            # Booleano auxiliar para la poda
            continuar: bool = True
            # Calculo de cabezas y colas
            continuar = nodo_act.cabezas_colas(mejor_solucion[1])
            if not continuar:
                continue
            # Metodo para fijar disyunciones adicionales
            continuar = nodo_act.metodo_adicional(mejor_solucion[1])
            if not continuar:
                continue
            # Calculo de la solucion heurustica, su camino critico y el peso de este
            sol_heur: Nodo = nodo_act.sol_heur()
            camino_critico: List[Operacion] = sol_heur.camino_critico()
            sol_heur_cmax: int = sol_heur.peso_camino(camino_critico)
            # Si la solucion obtenida es mejor actualizamos
            if sol_heur_cmax < mejor_solucion[1]:
                mejor_solucion[0] = sol_heur
                mejor_solucion[1] = sol_heur_cmax
                print(sol_heur_cmax)
            # Calculo de los bloques
            bloques: List[List[Operacion]] = nodo_act.bloques(camino_critico)
            # Si no hay mejores soluciones
            if len(bloques) == 0:
                return
            # Calculo de los conjuntos de ramificacion
            conjuntos_ramificacion: List[Tuple[int, int, List[Operacion]]] = nodo_act.conjuntos_ramificaion(bloques, camino_critico)
            # ED adicional para insertar los nodos en orden en la frontera
            frontera_parcial: List[Nodo] = []
            # Comenzamos la ramificacion
            for conjunto in conjuntos_ramificacion:
                # Ordenamos las operaciones
                op_ordenadas: List[Operacion]
                if conjunto[1] == 0:
                    op_ordenadas = sorted(conjunto[2], key = nodo_act.orden_cabeza)
                if conjunto[1] == 1:
                    op_ordenadas = sorted(conjunto[2], key = nodo_act.orden_cola)
                # Creamos los sucesores
                sucesor: Nodo
                for op in op_ordenadas:
                    # Si es Before
                    if conjunto[1] == 0:
                        sucesor = copy.deepcopy(nodo_act)
                        for op_bloque in bloques[conjunto[0]]:
                            if op_bloque != op:
                                # Las fijamos en el sucesor
                                sucesor.operaciones()[op.id()].anhade_sucesor_maquina(sucesor.operaciones()[op_bloque.id()])
                                sucesor.operaciones()[op_bloque.id()].anhade_predecesor_maquina(sucesor.operaciones()[op.id()])
                    # Si es After
                    if conjunto[1] == 1:
                        sucesor = copy.deepcopy(nodo_act)
                        for op_bloque in bloques[conjunto[0]]:
                            if op_bloque != op:
                                # Las fijamos en el sucesor
                                sucesor.operaciones()[op.id()].anhade_predecesor_maquina(sucesor.operaciones()[op_bloque.id()])
                                sucesor.operaciones()[op_bloque.id()].anhade_sucesor_maquina(sucesor.operaciones()[op.id()])
                    # Comprobamos que cumplen la cota
                    if nodo_act.cota_inf(op, bloques[conjunto[0]], conjunto[1]) < mejor_solucion[1]:
                        frontera_parcial.append(sucesor)
                # Una vez inspeccionado el conjuto podemos fijar otras aristas
                # Si es Before
                if conjunto[1] == 0:
                    primera_op: Operacion = bloques[conjunto[0]][0]
                    for op_bloque in bloques[conjunto[0]]:
                        if primera_op != op_bloque:
                            # Las fijamos en el nodo. De esta forma las tienen los sucesores
                            nodo_act.operaciones()[primera_op.id()].anhade_sucesor_maquina(nodo_act.operaciones()[op_bloque.id()])
                            nodo_act.operaciones()[op_bloque.id()].anhade_predecesor_maquina(nodo_act.operaciones()[primera_op.id()])
                # Si es after
                if conjunto[1] == 1:
                    ultima_op: Operacion = bloques[conjunto[0]][-1]
                    for op_bloque in bloques[conjunto[0]]:
                        if ultima_op != op_bloque:
                            # Las fijamos en el nodo. De esta forma las tienen los sucesores
                            nodo_act.operaciones()[op_bloque.id()].anhade_sucesor_maquina(nodo_act.operaciones()[ultima_op.id()])
                            nodo_act.operaciones()[ultima_op.id()].anhade_predecesor_maquina(nodo_act.operaciones()[op_bloque.id()])
            # Insertamos los nodos en la frontera en orden
            while len(frontera_parcial) > 0:
                frontera.append(frontera_parcial.pop())
        # Retornamos la mejor solucion
        return mejor_solucion